C===================================================================
C
      PROGRAM TORSCONMU
C
C  This program solves the Saint-Venant TORSION problem 
C  for domains with (MU)ltiple boundaries 
C  as a Neumann problem for the Laplace equation using
C  the boundary element method with (CON)stant
C  boundary elements 
C     
      IMPLICIT REAL*8 (A-H,O-Z)                     
      CHARACTER*15 INPUTFILE,OUTPUTFILE
C
C  Set the maximum dimensions
C   
      PARAMETER (N=968) 
      PARAMETER (IN=592) !961) 
      PARAMETER (NB=2)
C               
C   N= Number of boundary elements which is equal to the number 
C      of boundary nodes
C  IN= Number of internal points where the function u is calculated
C  NB= Number of boundaries of the multiple boundary domain
C  NL= Array with entries the number of the last element on each
C      boundary
C
      DIMENSION INDEX(N),NL(NB)
      DIMENSION XL(N+1),YL(N+1),XM(N),YM(N),G(N,N),H(N,N)
      DIMENSION UB(N),A(N,N),UNB(N),XIN(IN+1),YIN(IN+1),UIN(IN+1)
	1              ,TTZ(N),IC(N),enx(n),eny(n)
C
C  Read the names and open the input and output files
C   
      WRITE (*,'(A)')' Name of the INPUTFILE (max.15 characters)'
      READ  (*,'(A)') INPUTFILE
      WRITE (*,'(A)')' Name of the OUTPUTFILE (max.15 characters)'
      READ  (*,'(A)') OUTPUTFILE
      OPEN (UNIT=1, FILE=INPUTFILE)
      OPEN (UNIT=2, FILE=OUTPUTFILE)  
C
C  Read data from INPUTFILE
C
      CALL INPUT (XL,YL,XM,YM,XIN,YIN,INDEX,UB,N,IN,NL,NB,IC,ENX,ENY)
C
C  Compute the G matrix
C
      CALL GMATR (XL,YL,G,N,NL,NB) 
C         
C  Compute the H matrix
C
      CALL HMATR (XL,YL,H,N,NL,NB)
C
C  Form the system of equations AX=B
C
      CALL ABMATR (G,H,A,UNB,UB,INDEX,N)
C
C  Solve the system of equations
C
      CALL SOLVEQ(A,UNB,N,LSING)
C      
C  Form the vectors U and UN of all the boundary values
C
      CALL REORDER (UB,UNB,INDEX,N)

	CALL TORCENTER (XL,YL,N,UB,UNB,XTC,YTC,CT,NL,NB)
C
C  Compute the values of U at the internal points
C
	CALL UINTER(XL,YL,XM,YM,XIN,YIN,UB,UNB,UIN,N,IN,XTC,YTC
	1    ,CT,NL,NB)
C
C  Compute the the torsion constant D
C
	CALL TORSTIF (XL,YL,N,UB,XTC,YTC,D,NL,NB)
C
C  Compute the boundary stress Ttz
C
	CALL TORSTRESS(XL,YL,XM,YM,UB,XTC,YTC,TTZ,SL,N,NL,NB,IC)

C  Print the results in the OUTPUTFILE
C
      CALL OUTPUT (XM,YM,UB,UNB,XIN,YIN,UIN,N,IN,TTZ,XTC,YTC
	1       ,CT,D)
C
C  Close input and output files
C
      CLOSE(1)
      CLOSE(2)
C
      STOP
      END
C
C      
C====================================================================
C
      SUBROUTINE INPUT (XL,YL,XM,YM,XIN,YIN,INDEX,UB,N,IN,NL,NB,IC
	1              ,enx,eny)   
C     
C  This subroutine reads the input data from the input file
C  and writes them in the output file 
C
      IMPLICIT REAL*8 (A-H,O-Z)
      CHARACTER*80 NAME,TITLE
      DIMENSION XL(N+1),YL(N+1),XIN(IN+1),YIN(IN+1),INDEX(N),UB(N)
	1          ,NL(NB),XM(N),YM(N),IC(N),enx(n),eny(n)
C
      WRITE(2,100)
  100 FORMAT('',69('*'))
C
C  Read user's name
C
      READ(1,'(A)')NAME
C
      WRITE(2,'(A)')NAME
C
C  Read the title of the program
C
      READ(1,'(A)')TITLE
C
      WRITE(2,'(A)')TITLE
C
C  Read the number of the last element on each boundary
C
      READ(1,*)(NL(I),I=1,NB)
C
      WRITE(2,200)N,IN,NB
  200 FORMAT(//'BASIC PARAMETERS'//2X,'NUMBER OF BOUNDARY ELEMENTS='
     1,I3/2X,'NUMBER OF INTERNAL POINTS WHERE THE FUNCTION IS CALCULATED
     1=',I3//2X,'NUMBER OF BOUNDARIES =',I3)
      DO 250 I=1,NB
  250 WRITE(2,260)I,NL(I)
  260 FORMAT(2X,'BOUNDARY',I3,2X,'LAST BOUNDARY ELEMENT=',I3)      
C
C  Read the coordinates XL,YL of the extreme points of the boundary elements
C
      READ(1,*) (XL(I),YL(I),ic(i),I=1,N)
C
C  Write the coordinates in the output file
C
      WRITE(2,300)
  300 FORMAT(//2X,'COORDINATES OF THE EXTREME POINTS OF THE BOUNDARY '
     1'ELEMENTS',//2X,'POINT',9X,'XL',15X,'YL')
      DO 20 I=1,N
   20 WRITE(2,400) I,XL(I),YL(I)
  400 FORMAT(2X,I3,2(3X,E14.5))
C      
C  Compute the boundary values of Un and store in UB(I) (I=1,N-1),UB(N)=0.
C
	ENX=0.
	ENY=0.
      DO 10 I=1,N-1
	I2=I+1
	I1=I
	DO J=1,NB
	 IF (I.EQ.NL(J)) THEN
	  IF (J.EQ.1) THEN
	    IA=0
	  ELSE 
	    IA=NL(J-1)
	  ENDIF
	  I2=1+IA
	  ENDIF
	ENDDO
 
      DX=XL(I2)-XL(I1)
      DY=YL(I2)-YL(I1)
      SL=DSQRT(DX**2+DY**2)
      ENX(i)=DY/SL
      ENY(i)=-DX/SL
      XM(I)=(XL(I1)+XL(I2))/2.
      YM(I)=(YL(I1)+YL(I2))/2.
      UB(I)=YM(I)*ENX(i)-XM(I)*ENY(i)
   10 INDEX(I)=1

	IF (NB.EQ.1) THEN
	XM(N)=(XL(N)+XL(NL(1)))/2.
	YM(N)=(YL(N)+YL(NL(1)))/2.
	ELSE
	XM(N)=(XL(N)+XL(NL(NB-1)+1))/2.
	YM(N)=(YL(N)+YL(NL(NB-1)+1))/2.
	ENDIF
      UB(N)=0.
      INDEX(N)=0

C  If INDEX(I)=0, the value UB(I) is the prescribed value of u. 
C  If INDEX(I)=1, the normal derivative un is prescribed.


C
C  Write the boundary conditions in the output file
C
      WRITE(2,500)
  500 FORMAT(//2X,'BOUNDARY CONDITIONS'//2X,'NODE',6X,'INDEX',
     1 7X,'PRESCRIBED VALUE')
      DO 30 I=1,N
   30 WRITE(2,600) I,INDEX(I),UB(I)
  600 FORMAT (2X,I3,9X,I1,8X,E14.5)
C
C  Read the coordinates of the internal points
C
      READ(1,*) (XIN(I),YIN(I),I=1,IN) 
      RETURN
      END
C      
C      
C==================================================================
C
      SUBROUTINE GMATR (XL,YL,G,N,NL,NB)
C
C  This subroutine computes the elements of the G matrix
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XL(N+1),YL(N+1),XM(N),YM(N),NL(NB)
      DIMENSION G(N,N)
C 
C  Compute the nodal coordinates and store them in the arrays XM 
C  and YM
C
	IF (NB.GT.1) GOTO 5
      XL(NL(1)+1)=XL(1)
      YL(NL(1)+1)=YL(1)
C
    5 DO 10  I=1,N
      XM(I)=(XL(I)+XL(I+1))/2.
   10 YM(I)=(YL(I)+YL(I+1))/2.                   
      IF(NB.LE.1) GO TO 40
      XM(NL(1))=(XL(NL(1))+XL(1))/2.
      YM(NL(1))=(YL(NL(1))+YL(1))/2.
      DO 15 K=2,NB
      XM(NL(K))=(XL(NL(K))+XL(NL(K-1)+1))/2. 
   15 YM(NL(K))=(YL(NL(K))+YL(NL(K-1)+1))/2.
C
C  Compute the elements of matrix G
C
   40 DO 20 I=1,N
      X0=XM(I)
      Y0=YM(I)
      DO 20 J=1,N
      X1=XL(J)
      Y1=YL(J)
      X2=XL(J+1)
      Y2=YL(J+1)      
      IF(NB.LE.1)GOTO 60
        IF(J.NE.NL(1))GOTO 50
          X2=XL(1)
          Y2=YL(1)
          GOTO 60
   50   DO 30 K=2,NB 
        IF(J.NE.NL(K))GOTO 30  
          X2=XL(NL(K-1)+1)
          Y2=YL(NL(K-1)+1)      
          GOTO 60       
   30   CONTINUE      
   60 IF(I.NE.J)THEN
        CALL RLINTC(X0,Y0,X1,Y1,X2,Y2,RESULT)     
        G(I,J)=RESULT
      ELSEIF(I.EQ.J)THEN
        CALL SLINTC(X1,Y1,X2,Y2,RESULT)
        G(I,J)=RESULT
      ENDIF
   20 CONTINUE
      RETURN
      END
C
C
C================================================================
C
      SUBROUTINE RLINTC(X0,Y0,X1,Y1,X2,Y2,RESULT)
C
C  This subroutine computes the off-diagonal elements of the 
C  matrix G 
C
C  RA= The distance of point O from the Gauss integration point 
C      on the boundary element
C
C  WG= The weights of the Gauss integration
C
C  XI= The coordinates of the Gauss integration points in the 
C      interval [-1,1]
C
C  XC,YC= The global coordinates of the Gauss integration points
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XC(4),YC(4),XI(4),WG(4)
      DATA XI/-0.86113631,-0.33998104,0.33998104,0.86113631/
      DATA WG/0.34785485,0.65214515,0.65214515,0.34785485/
      PI=ACOS(-1.)
      AX=(X2-X1)/2.
      AY=(Y2-Y1)/2.
      BX=(X2+X1)/2.
      BY=(Y2+Y1)/2.
C
C  Compute the line integral
C
      RESULT=0.
C
      DO 30 I=1,4
      XC(I)=AX*XI(I)+BX
      YC(I)=AY*XI(I)+BY
      RA=SQRT((XC(I)-X0)**2+(YC(I)-Y0)**2)
   30 RESULT=RESULT+DLOG(RA)*WG(I)
      SL=2.*SQRT(AX**2+AY**2)
      RESULT=RESULT*SL/(4.*PI)
      RETURN
      END
C
C
C=================================================================
C
      SUBROUTINE SLINTC(X1,Y1,X2,Y2,RESULT)
C
C  This subroutines computes the diagonal elements of the matrix G
C
      IMPLICIT REAL*8 (A-H,O-Z)
      PI=ACOS(-1.)
      AX=(X2-X1)/2.
      AY=(Y2-Y1)/2.
      SL=SQRT(AX**2+AY**2)
      RESULT=SL*(DLOG(SL)-1.)/PI
      RETURN
      END
C
C
C=================================================================
C
      SUBROUTINE HMATR(XL,YL,H,N,NL,NB)
C
C  This subroutine computes the elements of the H matrix
C
      IMPLICIT REAL*8 (A-H,O-Z)     
      DIMENSION XL(N+1),YL(N+1),XM(N),YM(N),NL(NB)
      DIMENSION H(N,N)
C
      PI=ACOS(-1.)
C
C  Compute the nodal coordinates and store them in the arrays XM 
C  and YM
C
	IF (NB.GT.1) GOTO 5
      XL(NL(1)+1)=XL(1)
      YL(NL(1)+1)=YL(1)
C
    5 DO 10  I=1,N
      XM(I)=(XL(I)+XL(I+1))/2.
   10 YM(I)=(YL(I)+YL(I+1))/2.                   
      IF(NB.LE.1) GO TO 40
      XM(NL(1))=(XL(NL(1))+XL(1))/2.
      YM(NL(1))=(YL(NL(1))+YL(1))/2.
      DO 15 K=2,NB
      XM(NL(K))=(XL(NL(K))+XL(NL(K-1)+1))/2. 
   15 YM(NL(K))=(YL(NL(K))+YL(NL(K-1)+1))/2.
C
C  Compute the elements of H matrix
C
   40 DO 20 I=1,N 
      X0=XM(I)
      Y0=YM(I)
      DO 20 J=1,N
      X1=XL(J)
      Y1=YL(J)
      X2=XL(J+1)
      Y2=YL(J+1)      
      IF(NB.LE.1)GOTO 60
        IF(J.NE.NL(1))GOTO 50
          X2=XL(1)
          Y2=YL(1)
          GOTO 60
   50   DO 30 K=2,NB 
        IF(J.NE.NL(K))GOTO 30  
          X2=XL(NL(K-1)+1)
          Y2=YL(NL(K-1)+1)      
          GOTO 60       
   30   CONTINUE      
   60 IF (I.NE.J) THEN      
        CALL DALPHA (X0,Y0,X1,Y1,X2,Y2,RESULT)
        H(I,J)=RESULT
      ELSEIF (I.EQ.J) THEN
        H(I,J)=-0.5
      ENDIF
   20 CONTINUE
      RETURN
      END
C
C
C=================================================================
C
      SUBROUTINE DALPHA(X0,Y0,X1,Y1,X2,Y2,RESULT)
C
C  This subroutine computes the off diagonal elements of 
C  the H matrix
C
      IMPLICIT REAL*8 (A-H,O-Z)
      PI=ACOS(-1.)
      DY1=Y1-Y0
      DX1=X1-X0
      DY2=Y2-Y0
      DX2=X2-X0
      DL1=SQRT(DX1**2+DY1**2)
      COS1=DX1/DL1
      SIN1=DY1/DL1
      DX2R=DX2*COS1+DY2*SIN1
      DY2R=-DX2*SIN1+DY2*COS1
      DA=ATAN2(DY2R,DX2R)
      RESULT=DA/(2.*PI)
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE ABMATR (G,H,A,UNB,UB,INDEX,N)
C
C  This subroutine rearranges the matrices G and H and produces 
C  the matrices A and B=UNB
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION G(N,N),H(N,N),A(N,N),UNB(N),UB(N),INDEX(N)
C
C  Reorder the columns of the system of equations and 
C  store them in A
C
      DO 40 J=1,N
      IF(INDEX(J).EQ.0) THEN
      DO 20 I=1,N
   20 A(I,J)=-G(I,J)
      ELSEIF(INDEX(J).NE.0) THEN
      DO 30 I=1,N
   30 A(I,J)=H(I,J)
      END IF
   40 CONTINUE
C
C  Compute the right hand side vector and store it in UNB
C
      DO 50 I=1,N
      UNB(I)=0.
      DO 60 J=1,N
      IF(INDEX(J).EQ.0) THEN
      UNB(I)=UNB(I)-H(I,J)*UB(J) 
      ELSEIF(INDEX(J).NE.0)THEN
      UNB(I)=UNB(I)+G(I,J)*UB(J)
      ENDIF
   60 CONTINUE
   50 CONTINUE
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE SOLVEQ(A,UNB,N,LSING)
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(N,N),UNB(N)
      CALL LEQS(A,UNB,N,LSING)
      IF(LSING.EQ.0)THEN
      WRITE(2,150)
  150 FORMAT(/,'',69('*')//2X'The system has been solved regularly'/)
      ELSEIF(LSING.EQ.1)THEN
      WRITE(2,170)
  170 FORMAT(/,'',69('*')//2X'The system is singular'/)
      ENDIF
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE LEQS(A,B,N,LSING)
C
C   This subroutine uses Gauss elimination to solve
C   a system of linear equations, [A]{X}={B}, where
C   A : One-dimensional array which contains the occasional row of
C       the two-dimensional array of the coefficients of the unknowns
C   B : One-dimensional array which contains the known coefficients
C   N : Integer denoting the number of the unknowns
C   LSING: Integer taking the values:
C         LSING = 0, if the system has been solved regularly
C         LSING = 1, if the system is singular
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(1),B(1)
C
      LSING=0
      JJ=-N
      DO 10 J=1,N
        JY=J+1
        JJ=JJ+N+1
        AMAX=0.0
        IHELP=JJ-J
      DO 20 I=J,N
        IJ=IHELP+I
        IF(ABS(AMAX)-ABS(A(IJ))) 30,20,20
   30      AMAX=A(IJ)
           IMAX=I
   20 CONTINUE
         IF(ABS(AMAX).EQ.0.) THEN
            LSING=1
            RETURN
         END IF
      I1=J+N*(J-2)
      IHELP=IMAX-J
      DO 40 K=J,N
        I1=I1+N
        I2=I1+IHELP
        ATEMP=A(I1)
        A(I1)=A(I2)
        A(I2)=ATEMP
   40 A(I1)=A(I1)/AMAX
      ATEMP=B(IMAX)
      B(IMAX)=B(J)
      B(J)=ATEMP/AMAX
         IF(J-N) 50,70,50
   50       IQS=N*(J-1)
            DO 10 IX=JY,N
              IXJ=IQS+IX
              IHELP=J-IX
              DO 60 JX=JY,N
                IJREF=N*(JX-1)+IX
                JJX=IJREF+IHELP
   60           A(IJREF)=A(IJREF)-(A(IXJ)*A(JJX))
   10 B(IX)=B(IX)-B(J)*A(IXJ)
   70    NY=N-1
      NN=N*N
      DO 80 J=1,NY
        I1=NN-J
        I2=N-J
        I3=N
      DO 80 K=1,J
        B(I2)=B(I2)-A(I1)*B(I3)
        I1=I1-N
   80   I3=I3-1
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE REORDER(UB,UNB,INDEX,N)
C
C  This subroutine rearranges the arrays UB and UNB in such a way
C  that all values of the function u are stored in UB while all
C  the values of the normal derivative un are stored in UNB
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION UB(N),UNB(N),INDEX(N)
C
      DO 20  I=1,N
      IF(INDEX(I))20,20,10
   10 CH=UB(I)
      UB(I)=UNB(I)
      UNB(I)=CH
   20 CONTINUE
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE UINTER1(XL,YL,XIN,YIN,UB,UNB,UIN,N,IN,NL,NB)
C
C  This subroutine computes the values of u at the internal points
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XL(N+1),YL(N+1),XIN(IN+1),YIN(IN+1),UB(N),UNB(N)
	1           ,UIN(IN+1)
      DIMENSION NL(NB)
C
C  Compute the values of u at the internal points
C
	IF (NB.GT.1) GOTO 5
      XL(NL(1)+1)=XL(1)
      YL(NL(1)+1)=YL(1)
C
   5  DO 10  KK=1,IN
      UIN(KK)=0.
      DO 20  J=1,N
      X0=XIN(KK)
      Y0=YIN(KK)
      X1=XL(J)
      Y1=YL(J)
      X2=XL(J+1)
      Y2=YL(J+1)
      IF(NB.LE.1)GOTO 60
        IF(J.NE.NL(1))GOTO 50
          X2=XL(1)
          Y2=YL(1)
          GOTO 60
   50   DO 30 K=2,NB 
        IF(J.NE.NL(K))GOTO 30  
          X2=XL(NL(K-1)+1)
          Y2=YL(NL(K-1)+1)      
          GOTO 60       
   30   CONTINUE
   60 CALL DALPHA(X0,Y0,X1,Y1,X2,Y2,RESH)
      CALL RLINTC(X0,Y0,X1,Y1,X2,Y2,RESG)
   20 UIN(KK)=UIN(KK)+RESH*UB(J)-RESG*UNB(J)
   10 CONTINUE
      RETURN
      END


C
C
C==================================================================
C
      SUBROUTINE TORCENTER (XL,YL,N,UB,UNB,XTC,YTC,CT,NL,NB)
C
C
C  This subroutine computes the coordinates of the twist 
C  center of the cross-section and the arbitrary constant
C
      IMPLICIT REAL*8(A-H,O-Z)
      DIMENSION XL(N+1),YL(N+1),UB(N),UNB(N),NL(NB)
      DIMENSION XC(4),YC(4),XI(4),WG(4)
      DIMENSION AA(3,3),BB(3)
C
C        N = Number of boundary elements
C    XL,YL = Coordinates of the extreme points of the
C             boundary elements
C  XTC,YTC = Coordinates of the twist center
C       CT = Arbitrary constant of the Neumann problem 
C       WG = Weights of the Gauss integration
C       XI = Coordinates of the Gauss integration points
C             in the interval [-1,1]
C    XC,YC = Global coordinates of the Gauss integration points
C
      DATA XI/-0.86113631,-0.33998104,0.33998104,0.86113631/
      DATA WG/0.34785485,0.65214515,0.65214515,0.34785485/
      PI=ACOS(-1.)
C
      AREA=0.
      SX=0.
      SY=0.
      AIX=0.
      AIY=0.
      AIXY=0.
      AI1=0.
      AI2=0.
      AI3=0.
C
      DO 10 I=1,N
	I2=I+1
	I1=I
	do J=1,NB
	 IF (I.EQ.NL(J)) THEN
	  IF (J.EQ.1) THEN
	    IA=0
	  ELSE 
	    IA=NL(J-1)
	  ENDIF
	  I2=1+IA
	  ENDIF
	ENDDO



      AX=(XL(I2)-XL(I1))/2.
      AY=(YL(I2)-YL(I1))/2.
      BX=(XL(I2)+XL(I1))/2.
      BY=(YL(I2)+YL(I1))/2.

      SL=DSQRT(AX**2+AY**2)
      ENX=AY/SL
      ENY=-AX/SL
      TERMA=0.
      TERMSX=0.
      TERMSY=0.
      TERMIX=0.
      TERMIY=0.
      TERMIXY=0.
      TERMI1=0.
      TERMI2=0.
      TERMI3=0.
        DO 40 K=1,4
        XC(K)=AX*XI(K)+BX
        YC(K)=AY*XI(K)+BY
        TERMA=TERMA+0.5*(XC(K)*ENX+YC(K)*ENY)*WG(K)*SL
        TERMSX=TERMSX+0.5*YC(K)**2*ENY*WG(K)*SL 
        TERMSY=TERMSY+0.5*XC(K)**2*ENX*WG(K)*SL
        TERMIX=TERMIX+1./3.*YC(K)**3*ENY*WG(K)*SL
        TERMIY=TERMIY+1./3.*XC(K)**3*ENX*WG(K)*SL
        TERMIXY=TERMIXY+0.25*XC(K)*YC(K)*(XC(K)*ENX+YC(K)*ENY)
     1                  *WG(K)*SL
      U1=(XC(K)**2*YC(K)+YC(K)**3)/8.
      U1N=(2.*XC(K)*YC(K)*ENX+(XC(K)**2+3.*YC(K)**2)*ENY)/8.
        TERMI1=TERMI1+(UB(I)*U1N-UNB(I)*U1)*WG(K)*SL
      U2=(XC(K)**3+XC(K)*YC(K)**2)/8.
      U2N=((3.*XC(K)**2+YC(K)**2)*ENX+2.*XC(K)*YC(K)*ENY)/8.
        TERMI2=TERMI2+(UB(I)*U2N-UNB(I)*U2)*WG(K)*SL
      U3=(XC(K)**2+YC(K)**2)/4.
      U3N=(XC(K)*ENX+YC(K)*ENY)/2.
        TERMI3=TERMI3+(UB(I)*U3N-UNB(I)*U3)*WG(K)*SL
   40   CONTINUE
      AREA=AREA+TERMA
      SX=SX+TERMSX
      SY=SY+TERMSY
      AIX=AIX+TERMIX
      AIY=AIY+TERMIY
      AIXY=AIXY+TERMIXY
      AI1=AI1+TERMI1
      AI2=AI2+TERMI2
      AI3=AI3+TERMI3
   10 CONTINUE
C
C  Coordinates of the twist center and Neumann's constant
C
      AA(1,1)=AIX
      AA(1,2)=-AIXY
      AA(1,3)=-SX
      AA(2,1)=-AIXY
      AA(2,2)=AIY
      AA(2,3)=SY
      AA(3,1)=SX
      AA(3,2)=-SY
      AA(3,3)=-AREA
 
      BB(1)=-AI1
      BB(2)=AI2
      BB(3)=-AI3
      CALL LEQS(AA,BB,3,LSING)
      XTC=BB(1)
      YTC=BB(2)
      CT=BB(3)
C
      RETURN
      END      
C
C==================================================================
C
      SUBROUTINE UINTER(XL,YL,XM,YM,XIN,YIN,UB,UNB,UIN,N,IN,XTC,YTC
	1    ,CT,NL,NB)
C
C
C  This subroutine computes the values of u at the internal points
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XL(N+1),YL(N+1),XIN(IN+1),YIN(IN+1),NL(NB)
      DIMENSION UB(N),UNB(N),UIN(IN+1),XM(N),YM(N)
C
      XIN(IN+1)=XTC
      YIN(IN+1)=YTC
C
C  Compute the values of u at the internal points
C
      DO 10  K=1,IN+1
      UIN(K)=0.
      DO 20  J=1,N

	JP1=J+1

	do I=1,NB
	 IF (J.EQ.NL(I)) THEN
	  IF (I.EQ.1) THEN
	    IA=0
	  ELSE 
	    IA=NL(I-1)
	  ENDIF
	  JP1=1+IA
	  ENDIF
	ENDDO
	
      CALL DALPHA(XIN(K),YIN(K),XL(J),YL(J),XL(JP1),YL(JP1),RESH)
      CALL RLINTC(XIN(K),YIN(K),XL(J),YL(J),XL(JP1),YL(JP1),RESG)
   20 UIN(K)=UIN(K)+RESH*UB(J)-RESG*UNB(J)
   10 CONTINUE
C
      DO 30 K=1,N

	I2=k+1
 	I1=k
	do J=1,NB
	 IF (k.EQ.NL(J)) THEN
	  IF (J.EQ.1) THEN
	    IA=0
	  ELSE 
	    IA=NL(J-1)
	  ENDIF
	  I2=1+IA
	  ENDIF
	ENDDO

      DX=XL(I2)-XL(I1)
      DY=YL(I2)-YL(I1)

      SL=DSQRT(DX**2+DY**2)
      ENX=DY/SL
      ENY=-DX/SL
      UB(K)=UB(K)-YTC*XM(K)+XTC*YM(K)-CT
      UNB(K)=UNB(K)-YTC*ENX+XTC*ENY
   30 CONTINUE
C
      DO 40 K=1,IN+1
      UIN(K)=UIN(K)-YTC*XIN(K)+XTC*YIN(K)-CT
   40 CONTINUE  
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE TORSTIF (XL,YL,N,UB,XTC,YTC,D,NL,NB)
C
C  This subroutine computes the torsion constant
C
      IMPLICIT REAL*8 (A-H, O-Z)
      DIMENSION XL(N+1),YL(N+1),UB(N),NL(NB)
      DIMENSION XC(4),YC(4),XI(4),WG(4)
C
C      N = Number of boundary elements
C  XL,YL = Coordinates of the extreme points of the boundary elements
C      D = Torsion constant
C     WG = Weights of the Gauss integration
C     XI = Coordinates of the Gauss integration points in the
C          interval [-1, 1]
C  XC,YC = Global coordinates of the Gauss integration points
C
      DATA XI/-0.86113631,-0.33998104,0.33998104,0.86113631/
      DATA WG/0.34785485,0.65214515,0.65214515,0.34785485/
      D=0.
C
      DO 10 I=1,N

	I2=I+1
	I1=I
	do J=1,NB
	 IF (I.EQ.NL(J)) THEN
	  IF (J.EQ.1) THEN
	    IA=0
	  ELSE 
	    IA=NL(J-1)
	  ENDIF
	  I2=1+IA
	  ENDIF
	ENDDO

      AX=(XL(I2)-XL(I1))/2.
      AY=(YL(I2)-YL(I1))/2.
      BX=(XL(I2)+XL(I1))/2.
      BY=(YL(I2)+YL(I1))/2.

      SL=DSQRT(AX**2+AY**2)
      ENX=AY/SL
      ENY=-AX/SL
C
      TERM=0.
	TERM1=0.
        DO 40 K=1,4
        XC(K)=AX*XI(K)+BX-XTC
        YC(K)=AY*XI(K)+BY-YTC
        TERM=TERM+WG(K)*((XC(K)*YC(K)**2-YC(K)*UB(I))*ENX
     1           +(YC(K)*XC(K)**2+XC(K)*UB(I))*ENY)*SL

   40   CONTINUE
      D=D+TERM
   10 CONTINUE

	write (*,*) ' TORSIONAL CONSTANT D =',D

      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE TORSTRESS(XL,YL,XM,YM,UB,XTC,YTC,TTZ,SL,N,NL,NB,IC)
C
C
C  This subroutine computes the boundary shear stress tz 
C  in the tangential direction
C
C  TTZ = Shear stresses at the boundary nodal points
C   SL = Distances between the boundary nodal points
C
      IMPLICIT REAL*8 (A-H, O-Z)
      DIMENSION XL(N+1),YL(N+1),XM(N),YM(N)
      DIMENSION TTZ(N),UB(N),NL(NB),UBT(N)
	DIMENSION JN(N),KN(N),IC(N+1)

	DIMENSION TRM1(N,N),TRM(N,N),ENX(N),ENY(N)
	JN=0
  	KN=0
	TRM=0.

	DO I=1,N

	I2=I+1
	I1=I
	do J=1,NB
	 IF (I.EQ.NL(J)) THEN
	  IF (J.EQ.1) THEN
	    IA=0
	  ELSE 
	    IA=NL(J-1)
	  ENDIF
	  I2=1+IA
	  ENDIF
	ENDDO
	JN(I)=I1
	KN(I)=I2

      AX=(XL(I2)-XL(I1))/2.
      AY=(YL(I2)-YL(I1))/2.
      BX=(XL(I2)+XL(I1))/2.
      BY=(YL(I2)+YL(I1))/2.

      SL=DSQRT(AX**2+AY**2)
      ENX(I)=AY/SL
      ENY(I)=-AX/SL
	ENDDO
	TRM=0.
	NK=0
 	DO I=1,NB
	 IF (I.EQ.1) THEN
	  N1=1
	  N2=NL(1)
	  NM=N2-N1+1
	 ELSE
	  N1=NL(I-1)
	  N2=NL(I)
	  NM=N2-N1
	 ENDIF
	 TRM1=0.
	 call findif(XL,YL,IC,N1,N2,N,NM,TRM1,JN,KN)
	 DO J=1,NM
	  DO K=1,NM
	   TRM(J+NK,K+NK)=TRM1(J,K)
	  ENDDO
	 ENDDO
	 NK=NK+NM
	ENDDO
	UBT=0.
	UBT=MATMUL(TRM,UB)
	DO I=1,N
      TTZ(I)=UBT(I)+(XM(I)-XTC)*ENX(I)+(YM(I)-YTC)*ENY(I)
	ENDDO

      RETURN
      END
C
C
C================================================================
C
	SUBROUTINE FINDIF(XL,YL,ICN,N1,N2,N,NM,TRM,JN,KN)
C
C
C  This subroutine computes the finite difference matrix TRM
C  for multiply connected domains
C 
C  U,t=[TRM]*U
C

C  JN: array with the number of the initial point of each boundary
C      element
C  KN: array with the number of the end point of each boundary
C      element
C	 
	IMPLICIT REAL*8 (A-H,O-Z)                     
      DIMENSION XL(N+1),YL(N+1),TRM(N,N),ICN(N+1),JN(N),KN(N)

	TRM=0.
	DO I=1,NM
	X2=XL(N1+I-1)
	Y2=YL(N1+I-1)

      X3=XL(N1+I)
	Y3=YL(N1+I)
	sl2=0.
	SL2=DSQRT((X3-X2)**2+(Y3-Y2)**2)

	IF 	(ICN(N1+I-1).EQ.2) THEN

	X1=XL(N2)
	Y1=YL(N2)

	X2=XL(N1)
	Y2=YL(N1)
	
      X3=XL(N1+1)
	Y3=YL(N1+1)

	X4=XL(N1+2)
	Y4=YL(N1+2)
	sl1=0.
	sl3=0.
	SL1=DSQRT((X1-X2)**2+(Y1-Y2)**2)
	SL3=DSQRT((X3-X4)**2+(Y3-Y4)**2)
	sl2=0.
 	SL2=DSQRT((X3-X2)**2+(Y3-Y2)**2)

	S1=(sL1+SL2)/2.
	S2=(sL3+SL2)/2.
	A1=-S2/S1/(S1+S2)
	A2=-(S1-S2)/S1/S2
	A3=S1/S2/(S1+S2)
	TRM(I,nm)=A1
	TRM(I,I)=A2
	TRM(I,I+1)=A3

	GOTO 100

      
      elseif (ICN(N1+I-1).EQ.3) THEN
	X1=XL(N2-1)
	Y1=YL(N2-1)

	X2=XL(N2)
	Y2=YL(N2)

	X3=XL(N1)
 	Y3=YL(N1)

	X4=XL(N1+1)
	Y4=YL(N1+1)

	sl1=0.
	sl3=0.
	SL1=DSQRT((X1-X2)**2+(Y1-Y2)**2)
	SL3=DSQRT((X3-X4)**2+(Y3-Y4)**2)
	sl2=0.
 	SL2=DSQRT((X3-X2)**2+(Y3-Y2)**2)

	S1=(sL1+SL2)/2.
	S2=(sL3+SL2)/2.
	A1=-S2/S1/(S1+S2)
	A2=-(S1-S2)/S1/S2
	A3=S1/S2/(S1+S2)
	TRM(I,I-1)=A1
	TRM(I,I)=A2
	TRM(I,1)=A3

	GOTO 100
      
      ELSEIF (I.EQ.1.OR.ICN(N1+I-1).EQ.1) THEN

	X4=XL(N1+I+1)
 	Y4=YL(N1+I+1)

	X5=XL(N1+I+2)
	Y5=YL(N1+I+2)
	sl3=0.
	sl4=0.
	SL3=DSQRT((X3-X4)**2+(Y3-Y4)**2)
	SL4=DSQRT((X4-X5)**2+(Y4-Y5)**2)

	S1=(sL2+SL3)/2.
	S2=(sL3+SL4)/2.
	A1=-(2*S1+S2)/S1/(S1+S2)
	A2=(S1+S2)/S1/S2
	A3=-S1/S2/(S1+S2)

	TRM(I,I)=A1
	TRM(I,I+1)=A2
	TRM(I,I+2)=A3

	ELSEIF (I.EQ.Nm.OR.ICN(n1+I).EQ.1) THEN
	X0=XL(N1+I-3)
 	Y0=YL(N1+I-3)

	X1=XL(N1+I-2)
	Y1=YL(N1+I-2)
	sl0=0.
	sl1=0.
	SL0=DSQRT((X0-X1)**2+(Y0-Y1)**2)
	SL1=DSQRT((X1-X2)**2+(Y1-Y2)**2)

	S1=(sL2+SL1)/2.
	S2=(sL1+SL0)/2.
	A1=(2*S1+S2)/S1/(S1+S2)
	A2=-(S1+S2)/S1/S2
	A3=S1/S2/(S1+S2)

	TRM(I,I-2)=A3
	TRM(I,I-1)=A2
	TRM(I,I)=A1
	ELSE

	NI=N1+I-1
	X1=XL(JN(NI-1))
	Y1=YL(JN(NI-1))
	X2=XL(JN(NI))
	Y2=YL(JN(NI))
	X3=XL(KN(NI))
   	Y3=YL(KN(NI))
	X4=XL(KN(NI+1))
 	Y4=YL(KN(NI+1))


	sl1=0.
	sl3=0.
	SL1=DSQRT((X1-X2)**2+(Y1-Y2)**2)
	SL3=DSQRT((X3-X4)**2+(Y3-Y4)**2)
	sl2=0.
  	SL2=DSQRT((X3-X2)**2+(Y3-Y2)**2)

	S1=(sL1+SL2)/2.
	S2=(sL3+SL2)/2.
	A1=-S2/S1/(S1+S2)
	A2=-(S1-S2)/S1/S2
	A3=S1/S2/(S1+S2)
	TRM(I,I-1)=A1
	TRM(I,I)=A2
	TRM(I,I+1)=A3
100	ENDIF

	ENDDO

	RETURN
	END
C
C==================================================================
C
      SUBROUTINE OUTPUT(XM,YM,UB,UNB,XIN,YIN,UIN,N,IN,TTZ,XTC
	1     ,YTC,CT,D)
C
C  This subroutine prints the results in the output file.
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XM(N),YM(N),UB(N),UNB(N)
      DIMENSION XIN(IN+1),YIN(IN+1),UIN(IN+1),TTZ(N)
C
      WRITE(2,100)
  100 FORMAT('',69('*')//1X,'RESULTS'//2X,'BOUNDARY NODES'//
     1         11X,'X',15X,'Y',15X,'U',14X,'Un'/)
C
      DO 10 I=1,N
   10 WRITE(2,200) XM(I),YM(I),UB(I),UNB(I)
  200 FORMAT(4(2X,E14.5))
C
      WRITE (2,300)
  300 FORMAT(//,2X,'INTERNAL POINTS'//10X,'X',15X,'Y',11X,
     1 'SOLUTION U'/)
C
	DO 20 K=1,IN
   20 WRITE(2,400) XIN(K),YIN(K),UIN(K)
  400 FORMAT(3(2X,E14.5))

      WRITE(2,600) D
  600 FORMAT(/,'',69('*')//2X,'TORSION CONSTANT D=',E11.5)
      WRITE(2,500) 

      WRITE(2,700) 
  700 FORMAT(/,1X,'BOUNDARY STRESS Ttz'/)

	DO 30 I=1,N        
   30 WRITE(2,800) XM(I),YM(I),TTZ(I)
  800 FORMAT(3(2X,E14.5))

      WRITE(2,900)
  900 FORMAT(/'',69('*')//3X,'COORDINATES OF THE TWIST CENTER', 
     1       '   AND   ARBITRARY CONSTANT'/)
      WRITE(2,950)XTC,YTC,CT
  950 FORMAT(3X,'XTC=',E11.5,2X,'YTC=',E11.5,11X,'C=',E11.5)
      WRITE(2,500)

  500 FORMAT(/,'',69('*'))

      RETURN
      END          
C
C
C==================================================================      